/*
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (C) 2013 Intel Corp.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#ifndef _GRL_DLEYNA_SOURCE_H_
#define _GRL_DLEYNA_SOURCE_H_

#include <grilo.h>

#include "grl-dleyna-server.h"

#define GRL_DLEYNA_SOURCE_TYPE                          \
  (grl_dleyna_source_get_type ())

#define GRL_DLEYNA_SOURCE(obj)                          \
  (G_TYPE_CHECK_INSTANCE_CAST ((obj),                   \
                               GRL_DLEYNA_SOURCE_TYPE,  \
                               GrlDleynaSource))

#define GRL_IS_DLEYNA_SOURCE(obj)                       \
  (G_TYPE_CHECK_INSTANCE_TYPE ((obj),                   \
                               GRL_DLEYNA_SOURCE_TYPE))

#define GRL_DLEYNA_SOURCE_CLASS(klass)                  \
  (G_TYPE_CHECK_CLASS_CAST((klass),                     \
                           GRL_DLEYNA_SOURCE_TYPE,      \
                           GrlDleynaSourceClass))

#define GRL_IS_DLEYNA_SOURCE_CLASS(klass)               \
  (G_TYPE_CHECK_CLASS_TYPE((klass),                     \
                           GRL_DLEYNA_SOURCE_TYPE))

#define GRL_DLEYNA_SOURCE_GET_CLASS(obj)                \
  (G_TYPE_INSTANCE_GET_CLASS ((obj),                    \
                              GRL_DLEYNA_SOURCE_TYPE,   \
                              GrlDleynaSourceClass))

typedef struct _GrlDleynaSourcePrivate GrlDleynaSourcePrivate;
typedef struct _GrlDleynaSource GrlDleynaSource;

struct _GrlDleynaSource {
  GrlSource parent;

  /*< private >*/
  GrlDleynaSourcePrivate *priv;
};

typedef struct _GrlDleynaSourceClass GrlDleynaSourceClass;

struct _GrlDleynaSourceClass {
  GrlSourceClass parent_class;
};

GType             grl_dleyna_source_get_type (void);
gchar *           grl_dleyna_source_build_id (const gchar *udn);
GrlDleynaSource * grl_dleyna_source_new      (GrlDleynaServer *server);

#endif /* _GRL_DLEYNA_SOURCE_H_ */
