return {
  document = {
    article = {
      [1] = {
        p = {
          xml = "This is the first paragraph."
        },
        h2 = {
          xml = "Title with opt style",
          class = "opt"
        }
      },
      [2] = {
        p = {
          xml = "Some  text.",
          b = {
            xml = "important"
          }
        },
      }
    }
  }
}
