/*
 * Copyright (C) 2013 Victor Toso.
 *
 * Contact: Victor Toso <me@victortoso.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#ifndef _GRL_LUA_FACTORY_SOURCE_H_
#define _GRL_LUA_FACTORY_SOURCE_H_

#include <grilo.h>

#define GRL_LUA_FACTORY_SOURCE_TYPE                         (grl_lua_factory_source_get_type())
#define GRL_LUA_FACTORY_SOURCE(obj)                         (G_TYPE_CHECK_INSTANCE_CAST((obj), GRL_LUA_FACTORY_SOURCE_TYPE, GrlLuaFactorySource))
#define GRL_LUA_FACTORY_SOURCE_CLASS(klass)                 (G_TYPE_CHECK_CLASS_CAST((klass), GRL_LUA_FACTORY_SOURCE_TYPE, GrlLuaFactorySourceClass))
#define GRL_IS_LUA_FACTORY_SOURCE_CLASS(klass)              (G_TYPE_CHECK_CLASS_TYPE((klass), GRL_LUA_FACTORY_SOURCE_TYPE))
#define GRL_LUA_FACTORY_SOURCE_GET_CLASS(obj)               (G_TYPE_INSTANCE_GET_CLASS((obj), GRL_LUA_FACTORY_SOURCE_TYPE, GrlLuaFactorySourceClass))

typedef struct _GrlLuaFactorySource GrlLuaFactorySource;

typedef struct _GrlLuaFactorySourcePrivate GrlLuaFactorySourcePrivate;

struct _GrlLuaFactorySource {
  GrlSource parent;
  GrlLuaFactorySourcePrivate *priv;
};

typedef struct _GrlLuaFactorySourceClass GrlLuaFactorySourceClass;

struct _GrlLuaFactorySourceClass {
  GrlSourceClass parent_class;
};

GType grl_lua_factory_source_get_type (void);

#endif /* _GRL_LUA_FACTORY_SOURCE_H_ */
